use std::{any::Any, sync::Arc};

use async_trait::async_trait;
use chrono::{DateTime, Utc};
use sqlx::{Error, Pool, Postgres};

use crate::{
    adapter::gateway::post_db_service::PostDbService,
    application::error::post_error::PostError,
    domain::entity::{post::Post, post_info::PostInfo},
};

pub struct PostDbServiceImpl {
    db_pool: Arc<Pool<Postgres>>,
}

impl PostDbServiceImpl {
    pub fn new(db_pool: Arc<Pool<Postgres>>) -> Self {
        Self { db_pool }
    }
}

#[async_trait]
impl PostDbService for PostDbServiceImpl {
    async fn get_all_post_info(&self) -> Result<Vec<PostInfo>, PostError> {
        let post_records = sqlx::query!(
            r#"
            SELECT 
                p.id, 
                p.title, 
                p.description, 
                p.preview_image_url,
                p.published_time
            FROM post p
            WHERE p.deleted_time IS NULL
            "#
        )
        .fetch_all(&*self.db_pool)
        .await
        .map_err(|err| PostError::DatabaseError(err.to_string()))?;

        let mut post_infos = Vec::<PostInfo>::new();

        for post_record in post_records {
            let label_records = sqlx::query!(
                r#"
                SELECT 
                    l.id,
                    l.name,
                    l.color
                FROM label l
                JOIN post_label pl ON l.id = pl.label_id
                WHERE pl.post_id = $1
                AND l.deleted_time IS NULL
                "#,
                post_record.id
            )
            .fetch_all(&*self.db_pool)
            .await
            .map_err(|err| PostError::DatabaseError(err.to_string()))?;

            let labels = label_records
                .into_iter()
                .map(|label| crate::domain::entity::label::Label {
                    id: label.id,
                    name: label.name,
                    color: label.color as u32,
                })
                .collect();

            post_infos.push(PostInfo {
                id: post_record.id,
                title: post_record.title,
                description: post_record.description,
                preview_image_url: post_record.preview_image_url,
                labels: labels,
                published_time: post_record
                    .published_time
                    .map(|dt| DateTime::<Utc>::from_naive_utc_and_offset(dt, Utc)),
            });
        }

        Ok(post_infos)
    }

    async fn get_full_post(&self, id: i32) -> Result<Post, PostError> {
        let post_record = sqlx::query!(
            r#"
            SELECT
                p.id,
                p.title,
                p.description,
                p.preview_image_url,
                p.content,
                p.published_time
            FROM post p
            WHERE p.id = $1
            AND p.deleted_time IS NULL
            "#,
            id
        )
        .fetch_one(&*self.db_pool)
        .await
        .map_err(|err| {
            if err.type_id() == Error::RowNotFound.type_id() {
                PostError::NotFound
            } else {
                PostError::DatabaseError(err.to_string())
            }
        })?;

        // TODO: extract label selecting process
        let label_records = sqlx::query!(
            r#"
            SELECT 
                l.id,
                l.name,
                l.color
            FROM label l
            JOIN post_label pl ON l.id = pl.label_id
            WHERE pl.post_id = $1
            AND l.deleted_time IS NULL
            "#,
            post_record.id
        )
        .fetch_all(&*self.db_pool)
        .await
        .map_err(|err| PostError::DatabaseError(err.to_string()))?;

        let labels = label_records
            .into_iter()
            .map(|label| crate::domain::entity::label::Label {
                id: label.id,
                name: label.name,
                color: label.color as u32,
            })
            .collect();

        let post = Post {
            id: post_record.id,
            content: post_record.content,
            info: PostInfo {
                id: post_record.id,
                title: post_record.title,
                description: post_record.description,
                preview_image_url: post_record.preview_image_url,
                labels: labels,
                published_time: post_record
                    .published_time
                    .map(|dt| DateTime::<Utc>::from_naive_utc_and_offset(dt, Utc)),
            },
        };

        Ok(post)
    }
}
